from typing import List

from pyspark.sql.datasource import (
    DataSource,
    DataSourceReader,
    DataSourceStreamReader,
    InputPartition,
)
from pyspark.sql.types import StringType, StructType


def _validate_faker_schema(schema):
    # Verify the library is installed correctly.
    try:
        from faker import Faker
    except ImportError:
        raise Exception("You need to install `faker` to use the fake datasource.")

    fake = Faker()
    for field in schema.fields:
        try:
            getattr(fake, field.name)()
        except AttributeError:
            raise Exception(
                f"Unable to find a method called `{field.name}` in faker. "
                f"Please check Faker's documentation to see supported methods."
            )
        if field.dataType != StringType():
            raise Exception(
                f"Field `{field.name}` is not a StringType. "
                f"Only StringType is supported in the fake datasource."
            )


class FakeDataSource(DataSource):
    """
    A fake data source for PySpark to generate synthetic data using the `faker` library.

    This data source allows specifying a schema with field names that correspond to `faker`
    providers to generate random data for testing and development purposes.

    The default schema is `name string, date string, zipcode string, state string`, and the
    default number of rows is `3`. Both can be customized by users.

    Name: `fake`

    Notes
    -----
    - The fake data source relies on the `faker` library. Make sure it is installed and accessible.
    - Only string type fields are supported, and each field name must correspond to a method name in
      the `faker` library.
    - When using the stream reader, `numRows` is the number of rows per microbatch.

    Examples
    --------
    Register the data source.

    >>> from pyspark_datasources import FakeDataSource
    >>> spark.dataSource.register(FakeDataSource)

    Use the fake datasource with the default schema and default number of rows:

    >>> spark.read.format("fake").load().show()
    +-----------+----------+-------+-------+
    |       name|      date|zipcode|  state|
    +-----------+----------+-------+-------+
    |Carlos Cobb|2018-07-15|  73003|Indiana|
    | Eric Scott|1991-08-22|  10085|  Idaho|
    | Amy Martin|1988-10-28|  68076| Oregon|
    +-----------+----------+-------+-------+

    Use the fake datasource with a custom schema:

    >>> spark.read.format("fake").schema("name string, company string").load().show()
    +---------------------+--------------+
    |name                 |company       |
    +---------------------+--------------+
    |Tanner Brennan       |Adams Group   |
    |Leslie Maxwell       |Santiago Group|
    |Mrs. Jacqueline Brown|Maynard Inc   |
    +---------------------+--------------+

    Use the fake datasource with a different number of rows:

    >>> spark.read.format("fake").option("numRows", 5).load().show()
    +--------------+----------+-------+------------+
    |          name|      date|zipcode|       state|
    +--------------+----------+-------+------------+
    |  Pam Mitchell|1988-10-20|  23788|   Tennessee|
    |Melissa Turner|1996-06-14|  30851|      Nevada|
    |  Brian Ramsey|2021-08-21|  55277|  Washington|
    |  Caitlin Reed|1983-06-22|  89813|Pennsylvania|
    | Douglas James|2007-01-18|  46226|     Alabama|
    +--------------+----------+-------+------------+

    Streaming fake data:

    >>> stream = spark.readStream.format("fake").load().writeStream.format("console").start()
    Batch: 0
    +--------------+----------+-------+------------+
    |          name|      date|zipcode|       state|
    +--------------+----------+-------+------------+
    |    Tommy Diaz|1976-11-17|  27627|South Dakota|
    |Jonathan Perez|1986-02-23|  81307|Rhode Island|
    |  Julia Farmer|1990-10-10|  40482|    Virginia|
    +--------------+----------+-------+------------+
    Batch: 1
    ...
    >>> stream.stop()
    """

    @classmethod
    def name(cls):
        return "fake"

    def schema(self):
        return "name string, date string, zipcode string, state string"

    def reader(self, schema: StructType) -> "FakeDataSourceReader":
        _validate_faker_schema(schema)
        return FakeDataSourceReader(schema, self.options)

    def streamReader(self, schema) -> "FakeDataSourceStreamReader":
        _validate_faker_schema(schema)
        return FakeDataSourceStreamReader(schema, self.options)


class FakeDataSourceReader(DataSourceReader):

    def __init__(self, schema, options) -> None:
        self.schema: StructType = schema
        self.options = options

    def read(self, partition):
        from faker import Faker

        fake = Faker()
        # Note: every value in this `self.options` dictionary is a string.
        num_rows = int(self.options.get("numRows", 3))
        for _ in range(num_rows):
            row = []
            for field in self.schema.fields:
                value = getattr(fake, field.name)()
                row.append(value)
            yield tuple(row)


class FakeDataSourceStreamReader(DataSourceStreamReader):
    def __init__(self, schema, options) -> None:
        self.schema: StructType = schema
        self.rows_per_microbatch = int(options.get("numRows", 3))
        self.options = options
        self.offset = 0

    def initialOffset(self) -> dict:
        return {"offset": 0}

    def latestOffset(self) -> dict:
        self.offset += self.rows_per_microbatch
        return {"offset": self.offset}

    def partitions(self, start, end) -> List[InputPartition]:
        return [InputPartition(end["offset"] - start["offset"])]

    def read(self, partition):
        from faker import Faker

        fake = Faker()
        for _ in range(partition.value):
            row = []
            for field in self.schema.fields:
                value = getattr(fake, field.name)()
                row.append(value)
            yield tuple(row)



from pyspark.sql.datasource import DataSource, DataSourceReader
from pyspark.sql.types import StructType, StructField, StringType


class HuggingFaceDatasets(DataSource):
    """
    An example data source for reading HuggingFace Datasets in Spark.

    This data source allows reading public datasets from the HuggingFace Hub directly into Spark
    DataFrames. The schema is automatically inferred from the dataset features. The split can be
    specified using the `split` option. The default split is `train`.

    Name: `huggingface`

    Notes:
    -----
    - Please use the official HuggingFace Datasets API: https://github.com/huggingface/pyspark_huggingface.
    - The HuggingFace `datasets` library is required to use this data source. Make sure it is installed.
    - If the schema is automatically inferred, it will use string type for all fields.
    - Currently it can only be used with public datasets. Private or gated ones are not supported.

    Examples
    --------
    Register the data source.

    >>> from pyspark_datasources import HuggingFaceDatasets
    >>> spark.dataSource.register(HuggingFaceDatasets)

    Load a public dataset from the HuggingFace Hub.

    >>> spark.read.format("huggingface").load("imdb").show()
    +--------------------+-----+
    |                text|label|
    +--------------------+-----+
    |I rented I AM CUR...|    0|
    |"I Am Curious: Ye...|    0|
    |...                 |  ...|
    +--------------------+-----+

    Load a specific split from a public dataset from the HuggingFace Hub.

    >>> spark.read.format("huggingface").option("split", "test").load("imdb").show()
    +--------------------+-----+
    |                text|label|
    +--------------------+-----+
    |I love sci-fi and...|    0|
    |Worth the enterta...|    0|
    |...                 |  ...|
    +--------------------+-----+
    """

    def __init__(self, options):
        super().__init__(options)
        if "path" not in options or not options["path"]:
            raise Exception("You must specify a dataset name in`.load()`.")

    @classmethod
    def name(cls):
        return "huggingface"

    def schema(self):
        # The imports must be inside the method to be serializable.
        from datasets import load_dataset_builder
        dataset_name = self.options["path"]
        ds_builder = load_dataset_builder(dataset_name)
        features = ds_builder.info.features
        if features is None:
            raise Exception(
                "Unable to automatically determine the schema using the dataset features. "
                "Please specify the schema manually using `.schema()`."
            )
        schema = StructType()
        for key, value in features.items():
            # For simplicity, use string for all values.
            schema.add(StructField(key, StringType(), True))
        return schema

    def reader(self, schema: StructType) -> "DataSourceReader":
        return HuggingFaceDatasetsReader(schema, self.options)


class HuggingFaceDatasetsReader(DataSourceReader):
    def __init__(self, schema: StructType, options: dict):
        self.schema = schema
        self.dataset_name = options["path"]
        # TODO: validate the split value.
        self.split = options.get("split", "train")  # Default using train split.

    def read(self, partition):
        from datasets import load_dataset
        columns = [field.name for field in self.schema.fields]
        iter_dataset = load_dataset(self.dataset_name, split=self.split, streaming=True)
        for example in iter_dataset:
            yield tuple([example.get(column) for column in columns])



from dataclasses import dataclass
from typing import Dict, Optional

from pyspark.sql.datasource import DataSource, DataSourceReader
from pyspark.sql.types import StringType, StructField, StructType


@dataclass
class Sheet:
    """
    A dataclass to identify a Google Sheets document.

    Attributes
    ----------
    spreadsheet_id : str
        The ID of the Google Sheets document.
    sheet_id : str, optional
        The ID of the worksheet within the document.
    """

    spreadsheet_id: str
    sheet_id: Optional[str] = None  # if None, the first sheet is used

    @classmethod
    def from_url(cls, url: str) -> "Sheet":
        """
        Converts a Google Sheets URL to a Sheet object.
        """
        from urllib.parse import parse_qs, urlparse

        parsed = urlparse(url)
        if parsed.netloc != "docs.google.com" or not parsed.path.startswith(
            "/spreadsheets/d/"
        ):
            raise ValueError("URL is not a Google Sheets URL")
        qs = parse_qs(parsed.query)
        spreadsheet_id = parsed.path.split("/")[3]
        if "gid" in qs:
            sheet_id = qs["gid"][0]
        else:
            sheet_id = None
        return cls(spreadsheet_id, sheet_id)

    def get_query_url(self, query: Optional[str] = None):
        """
        Gets the query url that returns the results of the query as a CSV file.

        If no query is provided, returns the entire sheet.
        If sheet ID is None, uses the first sheet.

        See https://developers.google.com/chart/interactive/docs/querylanguage
        """
        from urllib.parse import urlencode

        path = f"https://docs.google.com/spreadsheets/d/{self.spreadsheet_id}/gviz/tq"
        url_query = {"tqx": "out:csv"}
        if self.sheet_id:
            url_query["gid"] = self.sheet_id
        if query:
            url_query["tq"] = query
        return f"{path}?{urlencode(url_query)}"


@dataclass
class Parameters:
    sheet: Sheet
    has_header: bool


class GoogleSheetsDataSource(DataSource):
    """
    A DataSource for reading table from public Google Sheets.

    Name: `googlesheets`

    Schema: By default, all columns are treated as strings and the header row defines the column names.

    Options
    --------
    - `url`: The URL of the Google Sheets document.
    - `path`: The ID of the Google Sheets document.
    - `sheet_id`: The ID of the worksheet within the document.
    - `has_header`: Whether the sheet has a header row. Default is `true`.

    Either `url` or `path` must be specified, but not both.

    Examples
    --------
    Register the data source.

    >>> from pyspark_datasources import GoogleSheetsDataSource
    >>> spark.dataSource.register(GoogleSheetsDataSource)

    Load data from a public Google Sheets document using `path` and optional `sheet_id`.

    >>> spreadsheet_id = "10pD8oRN3RTBJq976RKPWHuxYy0Qa_JOoGFpsaS0Lop0"
    >>> spark.read.format("googlesheets").options(sheet_id="0").load(spreadsheet_id).show()
    +-------+---------+---------+-------+
    |country| latitude|longitude|   name|
    +-------+---------+---------+-------+
    |     AD|42.546245| 1.601554|Andorra|
    |    ...|      ...|      ...|    ...|
    +-------+---------+---------+-------+

    Load data from a public Google Sheets document using `url`.

    >>> url = "https://docs.google.com/spreadsheets/d/10pD8oRN3RTBJq976RKPWHuxYy0Qa_JOoGFpsaS0Lop0/edit?gid=0#gid=0"
    >>> spark.read.format("googlesheets").options(url=url).load().show()
    +-------+---------+--------+-------+
    |country| latitude|ongitude|   name|
    +-------+---------+--------+-------+
    |     AD|42.546245|1.601554|Andorra|
    |    ...|      ...|     ...|    ...|
    +-------+---------+--------+-------+

    Specify custom schema.

    >>> schema = "id string, lat double, long double, name string"
    >>> spark.read.format("googlesheets").schema(schema).options(url=url).load().show()
    +---+---------+--------+-------+
    | id|      lat|    long|   name|
    +---+---------+--------+-------+
    | AD|42.546245|1.601554|Andorra|
    |...|      ...|     ...|    ...|
    +---+---------+--------+-------+

    Treat first row as data instead of header.

    >>> schema = "c1 string, c2 string, c3 string, c4 string"
    >>> spark.read.format("googlesheets").schema(schema).options(url=url, has_header="false").load().show()
    +-------+---------+---------+-------+
    |     c1|       c2|       c3|     c4|
    +-------+---------+---------+-------+
    |country| latitude|longitude|   name|
    |     AD|42.546245| 1.601554|Andorra|
    |    ...|      ...|      ...|    ...|
    +-------+---------+---------+-------+
    """

    @classmethod
    def name(self):
        return "googlesheets"

    def __init__(self, options: Dict[str, str]):
        if "url" in options:
            sheet = Sheet.from_url(options.pop("url"))
        elif "path" in options:
            sheet = Sheet(options.pop("path"), options.pop("sheet_id", None))
        else:
            raise ValueError(
                "You must specify either `url` or `path` (spreadsheet ID)."
            )
        has_header = options.pop("has_header", "true").lower() == "true"
        self.parameters = Parameters(sheet, has_header)

    def schema(self) -> StructType:
        if not self.parameters.has_header:
            raise ValueError("Custom schema is required when `has_header` is false")

        import pandas as pd

        # Read schema from the first row of the sheet
        df = pd.read_csv(self.parameters.sheet.get_query_url("select * limit 1"))
        return StructType([StructField(col, StringType()) for col in df.columns])

    def reader(self, schema: StructType) -> DataSourceReader:
        return GoogleSheetsReader(self.parameters, schema)


class GoogleSheetsReader(DataSourceReader):

    def __init__(self, parameters: Parameters, schema: StructType):
        self.parameters = parameters
        self.schema = schema

    def read(self, partition):
        from urllib.request import urlopen

        from pyarrow import csv
        from pyspark.sql.pandas.types import to_arrow_schema

        # Specify column types based on the schema
        convert_options = csv.ConvertOptions(
            column_types=to_arrow_schema(self.schema),
        )
        read_options = csv.ReadOptions(
            column_names=self.schema.fieldNames(),  # Rename columns
            skip_rows=(
                1 if self.parameters.has_header else 0  # Skip header row if present
            ),
        )
        with urlopen(self.parameters.sheet.get_query_url()) as file:
            yield from csv.read_csv(
                file, read_options=read_options, convert_options=convert_options
            ).to_batches()



import requests

from pyspark.sql import Row
from pyspark.sql.datasource import DataSource, DataSourceReader


class GithubDataSource(DataSource):
    """
    A DataSource for reading pull requests data from Github.

    Name: `github`

    Schema: `id int, title string, author string, created_at string, updated_at string`

    Examples
    --------
    Register the data source.

    >>> from pyspark_datasources import GithubDataSource
    >>> spark.dataSource.register(GithubDataSource)

    Load pull requests data from a public Github repository.

    >>> spark.read.format("github").load("apache/spark").show()
    +---+--------------------+--------+--------------------+--------------------+
    | id|               title|  author|          created_at|          updated_at|
    +---+--------------------+--------+--------------------+--------------------+
    |  1|Initial commit      |  matei |2014-02-03T18:47:...|2014-02-03T18:47:...|
    |...|                 ...|     ...|                 ...|                 ...|
    +---+--------------------+--------+--------------------+--------------------+

    Load pull requests data from a private Github repository.

    >>> spark.read.format("github").option("token", "your-token").load("owner/repo").show()
    """

    @classmethod
    def name(self):
        return "github"

    def schema(self):
        return "id int, title string, author string, created_at string, updated_at string"

    def reader(self, schema):
        return GithubPullRequestReader(self.options)


class GithubPullRequestReader(DataSourceReader):
    def __init__(self, options):
        self.token = options.get("token")
        self.repo = options.get("path")
        if self.repo is None:
            raise Exception(f"Must specify a repo in `.load()` method.")

    def read(self, partition):
        header = {
            "Accept": "application/vnd.github+json",
        }
        if self.token is not None:
            header["Authorization"] = f"Bearer {self.token}"
        url = f"https://api.github.com/repos/{self.repo}/pulls"
        response = requests.get(url)
        response.raise_for_status()
        prs = response.json()
        for pr in prs:
            yield Row(
                id = pr.get("number"),
                title = pr.get("title"),
                author = pr.get("user", {}).get("login"),
                created_at = pr.get("created_at"),
                updated_at = pr.get("updated_at")
            )




from dataclasses import dataclass
from typing import Dict

from pyspark.sql.types import StructType
from pyspark.sql.datasource import DataSource, DataSourceReader, InputPartition


class StockDataSource(DataSource):
    """
    A data source for reading stock data using the Alpha Vantage API.

    Examples
    --------

    Load the daily stock data for SPY:

    >>> df = spark.read.format("stock").option("api_key", "your-key").load("SPY")
    >>> df.show(n=5)
    +----------+------+------+------+------+--------+------+
    |      date|  open|  high|   low| close|  volume|symbol|
    +----------+------+------+------+------+--------+------+
    |2024-06-04|526.46|529.15|524.96|528.39|33898396|   SPY|
    |2024-06-03|529.02|529.31| 522.6| 527.8|46835702|   SPY|
    |2024-05-31|523.59| 527.5|518.36|527.37|90785755|   SPY|
    |2024-05-30|524.52| 525.2|521.33|522.61|46468510|   SPY|
    |2024-05-29|525.68|527.31|525.37| 526.1|45190323|   SPY|
    +----------+------+------+------+------+--------+------+
    """
    @classmethod
    def name(self) -> str:
        return "stock"

    def schema(self) -> str:
        return (
            "date string, open double, high double, "
            "low double, close double, volume long, symbol string"
        )

    def reader(self, schema):
        return StockDataReader(schema, self.options)


@dataclass
class Symbol(InputPartition):
    name: str


class StockDataReader(DataSourceReader):
    def __init__(self, schema: StructType, options: Dict):
        self.schema = schema
        self.options = options
        self.api_key = options.get("api_key")
        if not self.api_key:
            raise Exception(f"API Key is required to load the data.")
        # The name of the time-series. See https://www.alphavantage.co/documentation/ for more info.
        self.function = options.get("function", "TIME_SERIES_DAILY")
        if self.function not in ("TIME_SERIES_DAILY", "TIME_SERIES_WEEKLY", "TIME_SERIES_MONTHLY"):
            raise Exception(f"Function `{self.function}` is not supported.")

    def partitions(self):
        names = self.options["path"]
        # Split the names by comma and create a partition for each symbol.
        return [Symbol(name.strip()) for name in names.split(",")]

    def read(self, partition: Symbol):
        import requests
        symbol = partition.name
        resp = requests.get(
            f"https://www.alphavantage.co/query?"
            f"function={self.function}&symbol={symbol}&apikey={self.api_key}"
        )
        resp.raise_for_status()
        data = resp.json()
        key_name = next(key for key in data.keys() if key != "Meta Data")
        for date, info in data[key_name].items():
            yield (
                date,
                float(info["1. open"]),
                float(info["2. high"]),
                float(info["3. low"]),
                float(info["4. close"]),
                float(info["5. volume"]),
                symbol
            )



import io
import json
import time

from dataclasses import dataclass
from typing import Dict, List

from pyspark.sql.types import StructType
from pyspark.sql.datasource import DataSource, DataSourceWriter, WriterCommitMessage


class SimpleJsonDataSource(DataSource):
    """
    A simple json writer for writing data to Databricks DBFS.

    Examples
    --------

    >>> import pyspark.sql.functions as sf
    >>> df = spark.range(0, 10, 1, 2).withColumn("value", sf.expr("concat('value_', id)"))

    Register the data source.

    >>> from pyspark_datasources import SimpleJsonDataSource
    >>> spark.dataSource.register(SimpleJsonDataSource)

    Append the DataFrame to a DBFS path as json files.

    >>> (
    ...     df.write.format("simplejson")
    ...     .mode("append")
    ...     .option("databricks_url", "https://your-databricks-instance.cloud.databricks.com")
    ...     .option("databricks_token", "your-token")
    ...     .save("/path/to/output")
    ... )

    Overwrite the DataFrame to a DBFS path as json files.

    >>> (
    ...     df.write.format("simplejson")
    ...     .mode("overwrite")
    ...     .option("databricks_url", "https://your-databricks-instance.cloud.databricks.com")
    ...     .option("databricks_token", "your-token")
    ...     .save("/path/to/output")
    ... )
    """
    @classmethod
    def name(self) -> str:
        return "simplejson"

    def writer(self, schema: StructType, overwrite: bool):
        return SimpleJsonWriter(schema, self.options, overwrite)


@dataclass
class CommitMessage(WriterCommitMessage):
    output_path: str


class SimpleJsonWriter(DataSourceWriter):
    def __init__(self, schema: StructType, options: Dict, overwrite: bool):
        self.overwrite = overwrite
        self.databricks_url = options.get("databricks_url")
        self.databricks_token = options.get("databricks_token")
        if not self.databricks_url or not self.databricks_token:
            raise Exception("Databricks URL and token must be specified")
        self.path = options.get("path")
        if not self.path:
            raise Exception("You must specify an output path")

    def write(self, iterator):
        # Important: Always import non-serializable libraries inside the `write` method.
        from pyspark import TaskContext
        from databricks.sdk import WorkspaceClient

        # Consume all input rows and dump them as json.
        rows = [row.asDict() for row in iterator]
        json_data = json.dumps(rows)
        f = io.BytesIO(json_data.encode('utf-8'))

        context = TaskContext.get()
        id = context.taskAttemptId()
        file_path = f"{self.path}/{id}_{time.time_ns()}.json"

        # Upload to DFBS.
        w = WorkspaceClient(host=self.databricks_url, token=self.databricks_token)
        w.dbfs.upload(file_path, f)

        return CommitMessage(output_path=file_path)

    def commit(self, messages: List[CommitMessage]):
        from databricks.sdk import WorkspaceClient

        w = WorkspaceClient(host=self.databricks_url, token=self.databricks_token)
        paths = [message.output_path for message in messages]

        if self.overwrite:
            # Remove all files in the current directory except for the newly written files.
            for file in w.dbfs.list(self.path):
                if file.path not in paths:
                    print(f"[Overwrite] Removing file {file.path}")
                    w.dbfs.delete(file.path)

        # Write a success file
        file_path = f"{self.path}/_SUCCESS"
        f = io.BytesIO(b"success")
        w.dbfs.upload(file_path, f, overwrite=True)

    def abort(self, messages: List[CommitMessage]):
        from databricks.sdk import WorkspaceClient

        w = WorkspaceClient(host=self.databricks_url, token=self.databricks_token)
        # Clean up the newly written files
        for message in messages:
            if message is not None:
                print(f"[Abort] Removing up partially written files: {message.output_path}")
                w.dbfs.delete(message.output_path)



import ast
import json
import requests

from pyspark.sql.datasource import DataSource, SimpleDataSourceStreamReader
from pyspark.sql.types import StructType, StructField, DoubleType, StringType


class WeatherDataSource(DataSource):
    """
    A custom PySpark data source for fetching weather data from tomorrow.io for given 
    locations (latitude, longitude).

    Options
    -------

    - locations: specify a list of (latitude, longitude) tuples.
    - apikey: specify the API key for the weather service (tomorrow.io).
    - frequency: specify the frequency of the data ("minutely", "hourly", "daily"). 
                 Default is "minutely".

    Examples
    --------

    Register the data source.

    >>> from pyspark_datasources import WeatherDataSource
    >>> spark.dataSource.register(WeatherDataSource)


    Define the options for the custom data source
    
    >>> options = {
    ...    "locations": "[(37.7749, -122.4194), (40.7128, -74.0060)]",  # San Francisco and New York
    ...    "apikey": "your_api_key_here",
    ... }

    Create a DataFrame using the custom weather data source
    
    >>> weather_df = spark.readStream.format("weather").options(**options).load()

    Stream weather data and print the results to the console in real-time.

    >>> query = weather_df.writeStream.format("console").trigger(availableNow=True).start()
    """

    @classmethod
    def name(cls):
        """Returns the name of the data source."""
        return "weather"

    def __init__(self, options):
        """Initialize with options provided."""
        self.options = options
        self.frequency = options.get("frequency", "minutely")
        if self.frequency not in ["minutely", "hourly", "daily"]:
            raise ValueError(f"Unsupported frequency: {self.frequency}")

    def schema(self):
        """Defines the output schema of the data source."""
        return StructType([
            StructField("latitude", DoubleType(), True),
            StructField("longitude", DoubleType(), True),
            StructField("weather", StringType(), True),
            StructField("timestamp", StringType(), True),
        ])

    def simpleStreamReader(self, schema: StructType):
        """Returns an instance of the reader for this data source."""
        return WeatherSimpleStreamReader(schema, self.options)


class WeatherSimpleStreamReader(SimpleDataSourceStreamReader):

    def initialOffset(self):
        """
        Returns the initial offset for reading, which serves as the starting point for 
        the streaming data source.

        The initial offset is returned as a dictionary where each key is a unique identifier
        for a specific (latitude, longitude) pair, and each value is a timestamp string 
        (in ISO 8601 format) representing the point in time from which data should start being 
        read.

        Example:
            For locations [(37.7749, -122.4194), (40.7128, -74.0060)], the offset might look like:
            {
                "offset_37.7749_-122.4194": "2024-09-01T00:00:00Z",
                "offset_40.7128_-74.0060": "2024-09-01T00:00:00Z"
            }
        """
        return {f"offset_{lat}_{long}": "2024-09-01T00:00:00Z" for (lat, long) in self.locations}
    
    @staticmethod
    def _parse_locations(locations_str: str):
        """Converts string representation of list of tuples to actual list of tuples."""
        return [tuple(map(float, x)) for x in ast.literal_eval(locations_str)]

    def __init__(self, schema: StructType, options: dict):
        """Initialize with schema and options."""
        super().__init__()
        self.schema = schema
        self.locations = self._parse_locations(options.get("locations", "[]"))
        self.api_key = options.get("apikey", "")
        self.current = 0
        self.frequency = options.get("frequency", "minutely")
        self.session = requests.Session()  # Use a session for connection pooling

    def read(self, start: dict):
        """Reads data starting from the given offset."""
        data = []
        new_offset = {}
        for lat, long in self.locations:
            start_ts = start[f"offset_{lat}_{long}"]
            weather = self._fetch_weather(lat, long, self.api_key, self.session)[self.frequency]
            for entry in weather:
                # Start time is exclusive and end time is inclusive.
                if entry["time"] > start_ts:
                    data.append((lat, long, json.dumps(entry["values"]), entry["time"]))
            new_offset.update({f"offset_{lat}_{long}": weather[-1]["time"]})
        return (data, new_offset)

    @staticmethod
    def _fetch_weather(lat: float, long: float, api_key: str, session):
        """Fetches weather data for the given latitude and longitude using a REST API."""
        url = f"https://api.tomorrow.io/v4/weather/forecast?location={lat},{long}&apikey={api_key}"
        response = session.get(url)
        response.raise_for_status()
        return response.json()["timelines"]



from .fake import FakeDataSource
from .github import GithubDataSource
from .googlesheets import GoogleSheetsDataSource
from .huggingface import HuggingFaceDatasets
from .kaggle import KaggleDataSource
from .simplejson import SimpleJsonDataSource
from .stock import StockDataSource



import tempfile
from functools import cached_property
from typing import TYPE_CHECKING, Iterator

from pyspark.sql.datasource import DataSource, DataSourceReader
from pyspark.sql.pandas.types import from_arrow_schema
from pyspark.sql.types import StructType

if TYPE_CHECKING:
    import pyarrow as pa


class KaggleDataSource(DataSource):
    """
    A DataSource for reading Kaggle datasets in Spark.

    This data source allows reading datasets from Kaggle directly into Spark DataFrames.

    Name: `kaggle`

    Options
    -------
    - `handle`: The dataset handle on Kaggle, in the form of `{owner_slug}/{dataset_slug}`
        or `{owner_slug}/{dataset_slug}/versions/{version_number}`
    - `path`: The path to a file within the dataset.
    - `username`: The Kaggle username for authentication.
    - `key`: The Kaggle API key for authentication.

    Notes:
    -----
    - The `kagglehub` library is required to use this data source. Make sure it is installed.
    - To read private datasets or datasets that require user authentication, `username` and `key` must be provided.
    - Currently all data is read from a single partition.

    Examples
    --------
    Register the data source.

    >>> from pyspark_datasources import KaggleDataSource
    >>> spark.dataSource.register(KaggleDataSource)

    Load a public dataset from Kaggle.

    >>> spark.read.format("kaggle").options(handle="yasserh/titanic-dataset").load("Titanic-Dataset.csv").select("Name").show()
    +--------------------+
    |                Name|
    +--------------------+
    |Braund, Mr. Owen ...|
    |Cumings, Mrs. Joh...|
    |...                 |
    +--------------------+

    Load a private dataset with authentication.

    >>> spark.read.format("kaggle").options(
    ...     username="myaccount",
    ...     key="<token>",
    ...     handle="myaccount/my-private-dataset",
    ... ).load("file.csv").show()
    """

    @classmethod
    def name(cls) -> str:
        return "kaggle"

    @cached_property
    def _data(self) -> "pa.Table":
        import ast
        import os

        import pyarrow as pa

        handle = self.options.pop("handle")
        path = self.options.pop("path")
        username = self.options.pop("username", None)
        key = self.options.pop("key", None)
        if username or key:
            if not (username and key):
                raise ValueError(
                    "Both username and key must be provided to authenticate."
                )
            os.environ["KAGGLE_USERNAME"] = username
            os.environ["KAGGLE_KEY"] = key

        kwargs = {k: ast.literal_eval(v) for k, v in self.options.items()}

        # Cache in a temporary directory to avoid writing to ~ which may be read-only
        with tempfile.TemporaryDirectory() as tmpdir:
            os.environ["KAGGLEHUB_CACHE"] = tmpdir
            import kagglehub

            df = kagglehub.dataset_load(
                kagglehub.KaggleDatasetAdapter.PANDAS,
                handle,
                path,
                **kwargs,
            )
            return pa.Table.from_pandas(df)

    def schema(self) -> StructType:
        return from_arrow_schema(self._data.schema)

    def reader(self, schema: StructType) -> "KaggleDataReader":
        return KaggleDataReader(self)


class KaggleDataReader(DataSourceReader):
    def __init__(self, source: KaggleDataSource):
        self.source = source

    def read(self, partition) -> Iterator["pa.RecordBatch"]:
        yield from self.source._data.to_batches()



import lance
import pyarrow as pa

from dataclasses import dataclass
from typing import Iterator, List
from pyspark.sql.datasource import DataSource, DataSourceArrowWriter, WriterCommitMessage
from pyspark.sql.pandas.types import to_arrow_schema


class LanceSink(DataSource):
    """
    Write a Spark DataFrame into Lance format: https://lancedb.github.io/lance/index.html

    Note this requires Spark master branch nightly build to support `DataSourceArrowWriter`.

    Examples
    --------
    Register the data source:

    >>> from pyspark_datasources import LanceSink
    >>> spark.dataSource.register(LanceSink)

    Create a Spark dataframe with 2 partitions:

    >>> df = spark.createDataFrame([(1, "a"), (2, "b"), (3, "c")], schema="id int, value string")

    Save the dataframe in lance format:

    >>> df.write.format("lance").mode("append").save("/tmp/test_lance")
    /tmp/test_lance
    _transactions _versions     data

    Then you can use lance API to read the dataset:

    >>> import lance
    >>> ds = lance.LanceDataset("/tmp/test_lance")
    >>> ds.to_table().to_pandas()
       id
    0   0
    1   1
    2   2

    Notes
    -----
    - Currently this only works with Spark local mode. Cluster mode is not supported.
    """
    @classmethod
    def name(cls) -> str:
        return "lance"

    def writer(self, schema, overwrite: bool):
        if overwrite:
            raise Exception("Overwrite mode is not supported")
        if "path" not in self.options:
            raise Exception("Dataset URI must be specified when calling save()")
        return LanceWriter(schema, overwrite, self.options)


@dataclass
class LanceCommitMessage(WriterCommitMessage):
    fragments: List[lance.FragmentMetadata]


class LanceWriter(DataSourceArrowWriter):
    def __init__(self, schema, overwrite, options):
        self.options = options
        self.schema = schema  # Spark Schema (pyspark.sql.types.StructType)
        self.arrow_schema = to_arrow_schema(schema)  # Arrow schema (pa.StructType)
        self.uri = options["path"]
        assert not overwrite
        self.read_version = self._get_read_version()

    def _get_read_version(self):
        try:
            ds = lance.LanceDataset(self.uri)
            return ds.version
        except Exception:
            return None

    def write(self, iterator: Iterator[pa.RecordBatch]):
        reader = pa.RecordBatchReader.from_batches(self.arrow_schema, iterator)
        fragments = lance.fragment.write_fragments(reader, self.uri, schema=self.arrow_schema)
        return LanceCommitMessage(fragments=fragments)

    def commit(self, messages):
        fragments = [fragment for msg in messages for fragment in msg.fragments ]
        if self.read_version:
            # This means the dataset already exists.
            op = lance.LanceOperation.Append(fragments)
        else:
            # Create a new dataset.
            schema = to_arrow_schema(self.schema)
            op = lance.LanceOperation.Overwrite(schema, fragments)
        lance.LanceDataset.commit(self.uri, op, read_version=self.read_version)
